import cirq
import stimcirq


def test_mixture():
    r = stimcirq.TwoQubitAsymmetricDepolarizingChannel(
        [0.125, 0, 0, 0, 0, 0, 0.375, 0, 0, 0, 0, 0, 0, 0.25, 0]
    )
    assert r._dense_mixture_() == [
        (0.25, cirq.DensePauliString("II")),
        (0.125, cirq.DensePauliString("IX")),
        (0.375, cirq.DensePauliString("XZ")),
        (0.25, cirq.DensePauliString("ZY")),
    ]


def test_diagram():
    r = stimcirq.TwoQubitAsymmetricDepolarizingChannel(
        [0.125, 0, 0, 0, 0, 0, 0.375, 0, 0, 0, 0, 0, 0, 0.25, 0]
    )
    cirq.testing.assert_has_diagram(
        cirq.Circuit(r.on(*cirq.LineQubit.range(2))),
        """
0: ---PauliMix(II:0.25,IX:0.125,XZ:0.375,ZY:0.25)---
      |
1: ---#2--------------------------------------------
        """,
        use_unicode_characters=False,
    )


def test_repr():
    r = stimcirq.TwoQubitAsymmetricDepolarizingChannel(
        [0.125, 0, 0, 0, 0, 0, 0.375, 0, 0, 0, 0, 0, 0, 0.25, 0]
    )
    assert eval(repr(r), {'stimcirq': stimcirq}) == r


def test_json_serialization():
    r = stimcirq.TwoQubitAsymmetricDepolarizingChannel(
        [0.0125, 0.1, 0, 0.23, 0, 0, 0.0375, 0, 0, 0, 0, 0, 0, 0.25, 0]
    )
    c = cirq.Circuit(r(cirq.LineQubit(0), cirq.LineQubit(1)))
    json = cirq.to_json(c)
    c2 = cirq.read_json(json_text=json, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER])
    assert c == c2



import cirq
import numpy as np
import pytest
import stim
import stimcirq


def test_conversion():
    stim_circuit = stim.Circuit(
        """
        M 0
        DETECTOR(4) rec[-1]
        SHIFT_COORDS(1, 2, 3)
        DETECTOR(5) rec[-1]
        TICK
    """
    )
    cirq_circuit = cirq.Circuit(
        cirq.measure(cirq.LineQubit(0)),
        stimcirq.DetAnnotation(parity_keys=["0"], coordinate_metadata=[4]),
        stimcirq.ShiftCoordsAnnotation((1, 2, 3)),
        stimcirq.DetAnnotation(parity_keys=["0"], coordinate_metadata=[5]),
    )
    assert stimcirq.cirq_circuit_to_stim_circuit(cirq_circuit) == stim_circuit
    assert stimcirq.stim_circuit_to_cirq_circuit(stim_circuit) == cirq_circuit
    assert stimcirq.stim_circuit_to_cirq_circuit(stim_circuit, flatten=False) == cirq_circuit
    assert stimcirq.stim_circuit_to_cirq_circuit(stim_circuit, flatten=True) == cirq.Circuit(
        cirq.measure(cirq.LineQubit(0)),
        stimcirq.DetAnnotation(parity_keys=["0"], coordinate_metadata=[4]),
        stimcirq.DetAnnotation(parity_keys=["0"], coordinate_metadata=[6]),
    )


def test_simulation():
    a = cirq.LineQubit(0)
    s = cirq.Simulator().sample(
        cirq.Circuit(
            cirq.X(a), cirq.measure(a, key="a"), stimcirq.ShiftCoordsAnnotation((1, 2, 3))
        ),
        repetitions=3,
    )
    np.testing.assert_array_equal(s["a"], [1, 1, 1])


def test_diagram():
    a, b = cirq.LineQubit.range(2)
    cirq.testing.assert_has_diagram(
        cirq.Circuit(
            cirq.measure(a, key="a"),
            cirq.measure(b, key="b"),
            stimcirq.ShiftCoordsAnnotation([1, 2, 3]),
        ),
        """
0: ---M('a')---------------

1: ---M('b')---------------
      ShiftCoords(1,2,3)
        """,
        use_unicode_characters=False,
    )


def test_repr():
    val = stimcirq.ShiftCoordsAnnotation([1, 2, 3])
    assert eval(repr(val), {"stimcirq": stimcirq}) == val


def test_equality():
    eq = cirq.testing.EqualsTester()
    eq.add_equality_group(stimcirq.ShiftCoordsAnnotation([]))
    eq.add_equality_group(
        stimcirq.ShiftCoordsAnnotation([1, 2]), stimcirq.ShiftCoordsAnnotation((1, 2))
    )


def test_json_serialization():
    c = cirq.Circuit(stimcirq.ShiftCoordsAnnotation([1, 2, 3]))
    json = cirq.to_json(c)
    c2 = cirq.read_json(json_text=json, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER])
    assert c == c2



import cirq
import stimcirq


def test_repr():
    r = stimcirq.MeasureAndOrResetGate(
        measure=True,
        reset=True,
        basis='Y',
        key='result',
        invert_measure=True,
        measure_flip_probability=0.125,
    )
    assert eval(repr(r), {'stimcirq': stimcirq}) == r


def test_json_serialization():
    r = stimcirq.MeasureAndOrResetGate(
        measure=True,
        reset=True,
        basis='Y',
        key='result',
        invert_measure=True,
        measure_flip_probability=0.125,
    )
    c = cirq.Circuit(r(cirq.LineQubit(1)))
    json = cirq.to_json(c)
    c2 = cirq.read_json(json_text=json, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER])
    assert c == c2



from typing import Any, Dict, List, Sequence, Tuple

import cirq
import stim


@cirq.value_equality
class TwoQubitAsymmetricDepolarizingChannel(cirq.Gate):
    """A two qubit error channel that applies Pauli pairs according to the given probability
    distribution.

    The 15 probabilities are in IX, IY, ... ZY, ZZ order. II is skipped; it's the leftover.
    """

    def __init__(self, probabilities: Sequence[float]):
        if len(probabilities) != 15:
            raise ValueError(f"len(probabilities={probabilities}) != 15")
        if sum(probabilities) > 1.000001:
            raise ValueError(f"sum(probabilities={probabilities}) > 1")
        self.probabilities = tuple(probabilities)

    def _num_qubits_(self):
        return 2

    def _value_equality_values_(self):
        return self.probabilities

    def _has_mixture_(self):
        return True

    def _dense_mixture_(self):
        result = [(1 - sum(self.probabilities), cirq.DensePauliString([0, 0]))]
        result.extend(
            [
                (p, cirq.DensePauliString([((k + 1) >> 2) & 3, (k + 1) & 3]))
                for k, p in enumerate(self.probabilities)
            ]
        )
        return [(p, g) for p, g in result if p]

    def _mixture_(self):
        return [(p, cirq.unitary(g)) for p, g in self._dense_mixture_()]

    def _circuit_diagram_info_(self, args: cirq.CircuitDiagramInfoArgs):
        result = []
        for p, d in self._dense_mixture_():
            result.append(str(d)[1:] + ":" + args.format_real(p))
        return "PauliMix(" + ",".join(result) + ")", "#2"

    def _stim_conversion_(self, edit_circuit: stim.Circuit, targets: List[int], **kwargs):
        edit_circuit.append_operation("PAULI_CHANNEL_2", targets, self.probabilities)

    def __repr__(self):
        return f"stimcirq.TwoQubitAsymmetricDepolarizingChannel({self.probabilities!r})"

    def _json_dict_(self) -> Dict[str, Any]:
        return {'cirq_type': self.__class__.__name__, 'probabilities': list(self.probabilities)}



from ._cirq_to_stim import cirq_circuit_to_stim_circuit
from ._det_annotation import DetAnnotation
from ._obs_annotation import CumulativeObservableAnnotation
from ._shift_coords_annotation import ShiftCoordsAnnotation
from ._stim_sampler import StimSampler
from ._stim_to_cirq import (
    MeasureAndOrResetGate,
    stim_circuit_to_cirq_circuit,
    TwoQubitAsymmetricDepolarizingChannel,
)
from ._sweep_pauli import SweepPauli

JSON_RESOLVERS_DICT = {
    "CumulativeObservableAnnotation": CumulativeObservableAnnotation,
    "DetAnnotation": DetAnnotation,
    "MeasureAndOrResetGate": MeasureAndOrResetGate,
    "ShiftCoordsAnnotation": ShiftCoordsAnnotation,
    "SweepPauli": SweepPauli,
    "TwoQubitAsymmetricDepolarizingChannel": TwoQubitAsymmetricDepolarizingChannel,
}
JSON_RESOLVER = JSON_RESOLVERS_DICT.get

# Workaround for doctest not searching imported objects.
__test__ = {
    "cirq_circuit_to_stim_circuit": cirq_circuit_to_stim_circuit,
    "CumulativeObservableAnnotation": CumulativeObservableAnnotation,
    "DetAnnotation": DetAnnotation,
    "MeasureAndOrResetGate": MeasureAndOrResetGate,
    "ShiftCoordsAnnotation": ShiftCoordsAnnotation,
    "stim_circuit_to_cirq_circuit": stim_circuit_to_cirq_circuit,
    "StimSampler": StimSampler,
    "SweepPauli": SweepPauli,
    "TwoQubitAsymmetricDepolarizingChannel": TwoQubitAsymmetricDepolarizingChannel,
}



from typing import Any, Dict, List, Optional, Union

import cirq
import stim
import sympy


@cirq.value_equality
class SweepPauli(cirq.Gate):
    """A Pauli gate included depending on sweep data, with compatibility across stim and cirq.

    Includes a sweep bit index for stim and a sweep symbol for cirq.
    Creates sweep-controlled operations like `CX sweep[5] 0` when converting to a stim circuit.
    """

    def __init__(
        self,
        *,
        stim_sweep_bit_index: int,
        cirq_sweep_symbol: Optional[Union[str, sympy.Symbol]] = None,
        pauli: cirq.Pauli,
    ):
        r"""

        Args:
            stim_sweep_bit_index: The bit position, in some unspecified array, controlling the Pauli.
            cirq_sweep_symbol: The symbol used by cirq. Defaults to f"sweep_{sweep_bit_index}".
            pauli: The cirq Pauli operation to apply when the bit is True.
        """
        if cirq_sweep_symbol is None:
            cirq_sweep_symbol = f"sweep[{stim_sweep_bit_index}]"
        self.stim_sweep_bit_index = stim_sweep_bit_index
        self.cirq_sweep_symbol = cirq_sweep_symbol
        self.pauli = pauli

    def _decomposed_into_pauli(self) -> cirq.Gate:
        return self.pauli ** self.cirq_sweep_symbol

    def _resolve_parameters_(self, resolver: cirq.ParamResolver, recursive: bool) -> cirq.Gate:
        new_value = resolver.value_of(self.cirq_sweep_symbol, recursive=recursive)
        if new_value == self.cirq_sweep_symbol:
            return self
        return self.pauli ** new_value

    def _is_parameterized_(self) -> bool:
        return True

    def _decompose_(self, qubits) -> cirq.OP_TREE:
        return self.pauli.on(*qubits) ** self.cirq_sweep_symbol

    def _num_qubits_(self) -> int:
        return 1

    def _value_equality_values_(self) -> Any:
        return self.pauli, self.stim_sweep_bit_index, self.cirq_sweep_symbol

    def _circuit_diagram_info_(self, args: Any) -> str:
        return f"{self.pauli}^sweep[{self.stim_sweep_bit_index}]='{self.cirq_sweep_symbol}'"

    def _json_dict_(self) -> Dict[str, Any]:
        return {
            'cirq_type': self.__class__.__name__,
            'pauli': self.pauli,
            'stim_sweep_bit_index': self.stim_sweep_bit_index,
            'cirq_sweep_symbol': self.cirq_sweep_symbol,
        }

    def __repr__(self) -> str:
        return (
            f'stimcirq.SweepPauli('
            f'pauli={self.pauli!r}, '
            f'stim_sweep_bit_index={self.stim_sweep_bit_index!r}, '
            f'cirq_sweep_symbol={self.cirq_sweep_symbol!r})'
        )

    def _stim_conversion_(self, edit_circuit: stim.Circuit, targets: List[int], **kwargs):
        edit_circuit.append_operation(
            f"C{self.pauli}", [stim.target_sweep_bit(self.stim_sweep_bit_index)] + targets
        )



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

version = '1.8.0'

setup(
    name='stimcirq',
    version=version,
    author='Craig Gidney',
    author_email='craig.gidney@gmail.com',
    url='https://github.com/quantumlib/stim',
    license='Apache 2',
    packages=['stimcirq'],
    description='Implements a cirq.Sampler backed by stim.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    python_requires='>=3.6.0',
    data_files=['README.md'],
    install_requires=['stim', 'cirq-core'],
    tests_require=['pytest', 'python3-distutils'],
)



import cirq
import numpy as np
import pytest
import stim
import stimcirq


def test_stim_conversion():
    a, b, c = cirq.LineQubit.range(3)

    with pytest.raises(ValueError, match="earlier"):
        stimcirq.cirq_circuit_to_stim_circuit(
            cirq.Circuit(cirq.Moment(stimcirq.DetAnnotation(parity_keys=["unknown"])))
        )
    with pytest.raises(ValueError, match="earlier"):
        stimcirq.cirq_circuit_to_stim_circuit(
            cirq.Circuit(
                cirq.Moment(
                    stimcirq.DetAnnotation(parity_keys=["later"]), cirq.measure(b, key="later")
                )
            )
        )
    with pytest.raises(ValueError, match="earlier"):
        stimcirq.cirq_circuit_to_stim_circuit(
            cirq.Circuit(
                cirq.Moment(stimcirq.DetAnnotation(parity_keys=["later"])),
                cirq.Moment(cirq.measure(b, key="later")),
            )
        )
    assert stimcirq.cirq_circuit_to_stim_circuit(
        cirq.Circuit(
            cirq.Moment(cirq.measure(b, key="earlier")),
            cirq.Moment(stimcirq.DetAnnotation(parity_keys=["earlier"])),
        )
    ) == stim.Circuit(
        """
        QUBIT_COORDS(1) 0
        M 0
        TICK
        DETECTOR rec[-1]
        TICK
    """
    )
    assert stimcirq.cirq_circuit_to_stim_circuit(
        cirq.Circuit(
            cirq.Moment(
                cirq.measure(b, key="earlier"), stimcirq.DetAnnotation(parity_keys=["earlier"])
            )
        )
    ) == stim.Circuit(
        """
        QUBIT_COORDS(1) 0
        M 0
        DETECTOR rec[-1]
        TICK
    """
    )

    assert stimcirq.cirq_circuit_to_stim_circuit(
        cirq.Circuit(
            cirq.Moment(cirq.measure(a, key="a"), cirq.measure(b, key="b")),
            cirq.Moment(
                stimcirq.DetAnnotation(parity_keys=["a", "b"], coordinate_metadata=(1, 2, 3.5))
            ),
        )
    ) == stim.Circuit(
        """
        M 0 1
        TICK
        DETECTOR(1, 2, 3.5) rec[-1] rec[-2]
        TICK
    """
    )

    assert stimcirq.cirq_circuit_to_stim_circuit(
        cirq.Circuit(
            cirq.H(a),
            cirq.CNOT(a, b),
            cirq.CNOT(a, c),
            cirq.Moment(
                cirq.measure(a, key="a"),
                cirq.measure(b, key="b"),
                stimcirq.DetAnnotation(parity_keys=["a", "b"]),
                cirq.measure(c, key="c"),
                stimcirq.DetAnnotation(parity_keys=["a", "c"]),
            ),
        )
    ) == stim.Circuit(
        """
        H 0
        TICK
        CNOT 0 1
        TICK
        CNOT 0 2
        TICK
        M 0 1
        DETECTOR rec[-1] rec[-2]
        M 2
        DETECTOR rec[-1] rec[-3]
        TICK
    """
    )


def test_simulation():
    a = cirq.LineQubit(0)
    s = cirq.Simulator().sample(
        cirq.Circuit(
            cirq.X(a), cirq.measure(a, key="a"), stimcirq.DetAnnotation(parity_keys=["a"])
        ),
        repetitions=3,
    )
    np.testing.assert_array_equal(s["a"], [1, 1, 1])


def test_diagram():
    a, b = cirq.LineQubit.range(2)
    cirq.testing.assert_has_diagram(
        cirq.Circuit(
            cirq.measure(a, key="a"),
            cirq.measure(b, key="b"),
            stimcirq.DetAnnotation(parity_keys=["a", "b"]),
        ),
        """
0: ---M('a')---------

1: ---M('b')---------
      Det('a','b')
        """,
        use_unicode_characters=False,
    )


def test_repr():
    val = stimcirq.DetAnnotation(parity_keys=["a", "b"], coordinate_metadata=(1, 2))
    assert eval(repr(val), {"stimcirq": stimcirq}) == val


def test_equality():
    eq = cirq.testing.EqualsTester()
    eq.add_equality_group(stimcirq.DetAnnotation(), stimcirq.DetAnnotation())
    eq.add_equality_group(stimcirq.DetAnnotation(parity_keys=["a"]))
    eq.add_equality_group(
        stimcirq.DetAnnotation(parity_keys=["a"], coordinate_metadata=[1, 2]),
        stimcirq.DetAnnotation(parity_keys=["a"], coordinate_metadata=(1, 2)),
    )
    eq.add_equality_group(stimcirq.DetAnnotation(parity_keys=["a"], coordinate_metadata=(2, 1)))
    eq.add_equality_group(stimcirq.DetAnnotation(parity_keys=["b"]))
    eq.add_equality_group(
        stimcirq.DetAnnotation(parity_keys=["a", "b"]),
        stimcirq.DetAnnotation(parity_keys=["b", "a"]),
    )


def test_json_serialization():
    c = cirq.Circuit(
        stimcirq.DetAnnotation(parity_keys=["a", "b"]),
        stimcirq.DetAnnotation(parity_keys=["a", "b"], relative_keys=[-3, -5]),
        stimcirq.DetAnnotation(coordinate_metadata=(1, 2)),
        stimcirq.DetAnnotation(parity_keys=["d", "c"], coordinate_metadata=(1, 2, 3)),
    )
    json = cirq.to_json(c)
    c2 = cirq.read_json(json_text=json, resolvers=[*cirq.DEFAULT_RESOLVERS, stimcirq.JSON_RESOLVER])
    assert c == c2
